package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kbox2d.collision.e_nullProxy
import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kvector.util2d.AABB

sealed interface IFilter {
	/**
	 * The collision category bits. Normally you would just set one bit.
	 */
	val categoryBits: Long

	/**
	 * The collision mask bits. This states the categories that this
	 * shape would accept for collision.
	 */
	val maskBits: Long

	/**
	 * Collision groups allow a certain group of objects to never collide (negative)
	 * or always collide (positive). Zero means no collision group. Non-zero group
	 * filtering always wins against the mask bits.
	 */
	val groupIndex: Int
}

data class Filter(
	override var categoryBits: Long = 0x1L,
	override var maskBits: Long = 0xFFFFL,
	override var groupIndex: Int = 0,
) : IFilter {
	fun immutable(): ImmutableFilter {
		return ImmutableFilter(categoryBits, maskBits, groupIndex)
	}
}

data class ImmutableFilter(
	override val categoryBits: Long,
	override val maskBits: Long,
	override val groupIndex: Int,
) : IFilter

data class FixtureDef(
	/**
	 * The shape, this must be set. The shape will be cloned.
	 */
	var shape: IShape<*>? = null,

	var friction: Double = 0.2,

	/**
	 * The restitution (elasticity) usually in the range [0,1].
	 */
	var restitution: Double = 0.0,

	/**
	 * Restitution velocity threshold, usually in m/s. Collisions above this
	 * speed have restitution applied (will bounce).
	 */
	var restitutionThreshold: Double = 1.0 * b2_lengthUnitsPerMeter,

	/**
	 * The density, usually in kg/m^2.
	 */
	var density: Double = 0.0,

	/**
	 * A sensor shape collects contact information but never generates a collision
	 * response.
	 */
	var isSensor: Boolean = false,

	/**
	 * Use this to store application specific fixture data.
	 */
	var userData: Any? = null,

	val filter: Filter = Filter()
)

data class FixtureProxy(
	var aabb: AABB,
	val fixture: B2Fixture,
	val childIndex: Int,
) {
	private var setProxyID = false

	var proxyId: Int = e_nullProxy
		set(value) {
			if (!setProxyID) {
				field = value
				setProxyID = true
				return
			}

			throw IllegalStateException("FixtureProxy should be immutable (tried to set value $value, already having $field)")
		}
}
