
@file:Suppress("unused")

package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kvector.api.concrete.IMatrix2d
import ru.dbotthepony.kvector.api.concrete.IMatrix3d
import ru.dbotthepony.kvector.matrix.multiplyMatrix
import ru.dbotthepony.kvector.vector.ndouble.MutableVector2d
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.Vector3d
import ru.dbotthepony.kvector.vector.ndouble.cross
import kotlin.math.*

/// "Next Largest Power of 2
/// Given a binary integer value x, the next largest power of 2 can be computed by a SWAR algorithm
/// that recursively "folds" the upper bits into the lower bits. This process yields a bit vector with
/// the same most significant 1 as x, but all 1's below it. Adding 1 to that value yields the next
/// largest power of 2. For a 32-bit value:"
fun b2NextPowerOfTwo(x: Int): Int {
	var x = x
	x = x or (x ushr 1)
	x = x or (x ushr 2)
	x = x or (x ushr 4)
	x = x or (x ushr 8)
	x = x or (x ushr 16)
	return x + 1
}

fun b2IsPowerOfTwo(x: Int): Boolean {
	return x > 0 && (x and (x - 1)) == 0
}

/// Rotation
class Rotation(
	s: Double,
	c: Double,
) {
	/// Initialize from an angle in radians
	constructor(angle: Double) : this(sin(angle), cos(angle))

	override fun toString(): String {
		return "Rotation($s, $c)"
	}

	var s: Double = s
		set(value) {
			if (!value.isFinite()) {
				throw IllegalArgumentException("Tried to set illegal non-finite sinus $value")
			}

			if (value.isNaN()) {
				throw IllegalArgumentException("Tried to set illegal NaN sinus")
			}

			field = value
		}

	var c: Double = c
		set(value) {
			if (!value.isFinite()) {
				throw IllegalArgumentException("Tried to set illegal non-finite cosine $value")
			}

			if (value.isNaN()) {
				throw IllegalArgumentException("Tried to set illegal NaN cosine")
			}

			field = value
		}

	init {
		if (!s.isFinite()) {
			throw IllegalArgumentException("Sinus is infinite")
		}

		if (!c.isFinite()) {
			throw IllegalArgumentException("Cosines is infinite")
		}

		if (s.isNaN()) {
			throw IllegalArgumentException("Sinus is NaN")
		}

		if (c.isNaN()) {
			throw IllegalArgumentException("Cosines is NaN")
		}
	}

	/// Set to the identity rotation
	fun setIdentity() {
		s = 0.0
		c = 1.0
	}

	/// Set using an angle in radians.
	fun set(angle: Double) {
		s = sin(angle)
		c = cos(angle)
	}

	/// Get the angle in radians
	val angle: Double get() = atan2(s, c)

	/// Get the x-axis
	val xAxis: MutableVector2d get() = MutableVector2d(c, s)

	/// Get the u-axis
	val yAxis: MutableVector2d get() = MutableVector2d(-s, c)

	/// Multiply two rotations: q * r
	operator fun times(other: Rotation): Rotation {
		return Rotation(
			s = s * other.c + c * other.s,
			c = c * other.c - s * other.s,
		)
	}

	/// Transpose multiply two rotations: qT * r
	fun timesT(other: Rotation): Rotation {
		return Rotation(
			s = c * other.s - s * other.c,
			c = c * other.c + s * other.s,
		)
	}

	operator fun times(v: Vector2d): Vector2d {
		return Vector2d(c * v.x - s * v.y, s * v.x + c * v.y)
	}

	fun timesT(v: Vector2d): Vector2d {
		return Vector2d(c * v.x + s * v.y, -s * v.x + c * v.y)
	}
}

/**
 * A transform contains translation and rotation. It is used to represent
 * the position and orientation of rigid frames.
 */
class Transform(
	position: Vector2d,
	val rotation: Rotation,
) {
	var position: Vector2d = position
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal position $value")
			}

			field = value
		}

	override fun toString(): String {
		return "Transform($position, $rotation)"
	}

	constructor(position: Vector2d, rotation: Double) : this(position, Rotation(rotation))
	constructor() : this(Vector2d.ZERO, Rotation(0.0))

	// aliases for box2d C++ code
	val q by this::rotation
	var p by this::position

	/// Set this to the identity transform.
	fun setIdentity() {
		position = Vector2d.ZERO
		rotation.setIdentity()
	}

	/// Set this based on the position and angle.
	fun set(position: Vector2d, angle: Double) {
		this.position = Vector2d(position.x, position.y)
		rotation.set(angle)
	}

	operator fun times(other: Transform): Transform {
		return Transform(
			rotation = rotation * other.rotation,
			position = rotation.times(other.position) + position
		)
	}

	fun timesT(other: Transform): Transform {
		return Transform(
			rotation = rotation.timesT(other.rotation),
			position = rotation.timesT(other.position - position)
		)
	}

	operator fun times(v: Vector2d): Vector2d {
		return Vector2d(
			x = rotation.c * v.x - rotation.s * v.y + position.x,
			y = rotation.s * v.x + rotation.c * v.y + position.y
		)
	}

	fun timesT(v: Vector2d): Vector2d {
		val px = v.x - position.x
		val py = v.y - position.y

		return Vector2d(
			x = (rotation.c * px + rotation.s * py),
			y = (-rotation.s * px + rotation.c * py)
		)
	}
}

/**
 * This describes the motion of a body/shape for TOI computation.
 * Shapes are defined with respect to the body origin, which may
 * no coincide with the center of mass. However, to support dynamics
 * we must interpolate the center of mass position.
 */
class Sweep {
	/**
	 * local center of mass position
	 */
	var localCenter: Vector2d = Vector2d.ZERO
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal local center $value")
			}

			field = value
		}

	/**
	 * center world positions
	 */
	var c0: Vector2d = Vector2d.ZERO
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal center world position $value")
			}

			field = value
		}

	/**
	 * center world positions
	 */
	var c: Vector2d = Vector2d.ZERO
		set(value) {
			if (!value.isFinite) {
				throw IllegalArgumentException("Tried to set illegal center world position $value")
			}

			field = value
		}

	var oldCenter by this::c0
	var newCenter by this::c

	/**
	 * world angles
	 */
	var a0: Double = 0.0
		set(value) {
			if (!value.isFinite()) {
				throw IllegalArgumentException("Tried to set illegal non finite $value")
			}

			if (value.isNaN()) {
				throw IllegalArgumentException("Tried to set illegal NaN value")
			}

			field = value
		}

	/**
	 * world angles
	 */
	var a: Double = 0.0
		set(value) {
			if (!value.isFinite()) {
				throw IllegalArgumentException("Tried to set illegal non finite $value")
			}

			if (value.isNaN()) {
				throw IllegalArgumentException("Tried to set illegal NaN value")
			}

			field = value
		}

	var oldAngles by this::a0
	var newAngles by this::a

	/**
	 * Fraction of the current time step in the range [0,1]
	 * [c0] and [a0] are the positions at alpha0.
	 */
	var alpha0: Double = 0.0
		set(value) {
			if (!value.isFinite()) {
				throw IllegalArgumentException("Tried to set illegal non finite $value")
			}

			if (value.isNaN()) {
				throw IllegalArgumentException("Tried to set illegal NaN value")
			}

			field = value
		}

	/** Get the interpolated transform at a specific time.
	 * @param transform the output transform
	 * @param beta is a factor in [0,1], where 0 indicates [alpha0].
	 * https://fgiesen.wordpress.com/2012/08/15/linear-interpolation-past-present-and-future/
	 */
	fun getTransform(transform: Transform, beta: Double) {
		transform.position = c0 * (1.0 - beta) + c * beta
		val angle = (1.0 - beta) * a0 + beta * a
		transform.rotation.set(angle)
		transform.position -= transform.position * localCenter
	}

	/** Get the interpolated transform at a specific time.
	 * @param beta is a factor in [0,1], where 0 indicates [alpha0].
	 * https://fgiesen.wordpress.com/2012/08/15/linear-interpolation-past-present-and-future/
	 */
	fun getTransform(beta: Double): Transform {
		val v = Transform()
		getTransform(v, beta)
		return v
	}

	/**
	 * Advance the sweep forward, yielding a new initial state.
	 * @param alpha the new initial time.
	 */
	fun advance(alpha: Double) {
		require(alpha < 1.0) { "Bad advance value $alpha" }

		val beta = (alpha - alpha0) / (1.0 - alpha0)
		c0 += (c - c0) * beta
		a0 += (a - a0) * beta
		alpha0 = alpha
	}

	/**
	 * Normalize the angles.
	 */
	fun normalize() {
		val d = floor(a0 / (PI * 2.0)) * 2.0 * PI
		a0 -= d
		a -= d
	}

	fun copy(): Sweep {
		return Sweep().also {
			it.localCenter = localCenter
			it.c0 = c0
			it.c = c
			it.a0 = a0
			it.a = a
			it.alpha0 = alpha0
		}
	}

	fun load(from: Sweep) {
		this.localCenter = from.localCenter
		this.c0 = from.c0
		this.c = from.c
		this.a0 = from.a0
		this.a = from.a
		this.alpha0 = from.alpha0
	}
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Dot(a: Vector2d, b: Vector2d): Double {
	return a.dot(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Cross(a: Vector2d, b: Vector2d): Double {
	return a.cross(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Cross(a: Vector2d, b: Double): Vector2d {
	return a.cross(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Cross(a: Double, b: Vector2d): Vector2d {
	return a.cross(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Abs(a: Vector2d): Vector2d {
	return a.absoluteValue
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Abs(a: Double): Double {
	return a.absoluteValue
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Mul(q: Rotation, v: Vector2d): Vector2d {
	return q.times(v)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2MulT(q: Rotation, v: Vector2d): Vector2d {
	return q.timesT(v)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Mul(t: Transform, v: Vector2d): Vector2d {
	return t.times(v)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2MulT(t: Transform, v: Vector2d): Vector2d {
	return t.timesT(v)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Mul22(t: IMatrix3d<*>, v: Vector2d): Vector2d {
	val result = multiplyMatrix(t.toMatrix2d(), v)
	return Vector2d(result[0, 0], result[0, 1])
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Mul(t: Transform, v: Transform): Transform {
	return t.times(v)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2MulT(t: Transform, v: Transform): Transform {
	return t.timesT(v)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Mul(t: IMatrix2d<*>, v: Vector2d): Vector2d {
	val result = multiplyMatrix(t, v)
	return Vector2d(result[0, 0], result[0, 1])
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Mul(t: IMatrix3d<*>, v: Vector3d): Vector3d {
	val result = multiplyMatrix(t, v)
	return Vector3d(result[0, 0], result[0, 1], result[0, 2])
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Min(a: Double, b: Double): Double {
	return a.coerceAtMost(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Min(a: Vector2d, b: Vector2d): Vector2d {
	return a.coerceAtMost(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Max(a: Int, b: Int): Int {
	return a.coerceAtLeast(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Max(a: Double, b: Double): Double {
	return a.coerceAtLeast(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Max(a: Vector2d, b: Vector2d): Vector2d {
	return a.coerceAtLeast(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Distance(a: Vector2d, b: Vector2d): Double {
	return a.distance(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2DistanceSquared(a: Vector2d, b: Vector2d): Double {
	return a.distanceSquared(b)
}

/**
 * Shortcut for faster porting of C++ code
 */
@Suppress("NOTHING_TO_INLINE")
internal inline fun b2Clamp(a: Double, min: Double, max: Double): Double {
	return b2Max(min, b2Min(a, max))
}

private fun scalarDotWithCross(
	ax: Double,
	ay: Double,
	az: Double,

	bx: Double,
	by: Double,
	bz: Double,

	cx: Double,
	cy: Double,
	cz: Double,
): Double {
	val crossX = by * cz - bz * cy
	val crossY = bz * cx - bx * cz
	val crossZ = bx * cy - by * cx

	return ax * crossX + ay * crossY + az * crossZ
}

/**
 * Solve A * x = b, where b is a column vector. This is more efficient
 * than computing the inverse in one-shot cases.
 */
internal fun IMatrix3d<*>.solve(b: Vector3d): Vector3d {
	var determinant = determinant!!

	if (determinant != 0.0) {
		determinant = 1.0 / determinant
	}

	return Vector3d(
		x = determinant * scalarDotWithCross(
			b.x, b.y, b.z,
			r01, r11, r21,
			r02, r12, r22,
		),

		y = determinant * scalarDotWithCross(
			r00, r10, r20,
			b.x, b.y, b.z,
			r02, r12, r22,
		),

		z = determinant * scalarDotWithCross(
			r00, r10, r20,
			r01, r11, r21,
			b.x, b.y, b.z,
		),
	)
}

/**
 * Solve A * x = b, where b is a column vector. This is more efficient
 * than computing the inverse in one-shot cases.
 */
internal fun IMatrix3d<*>.solve(b: Vector2d): Vector2d {
	var determinant = determinant!!

	if (determinant != 0.0) {
		determinant = 1.0 / determinant
	}

	return Vector2d(
		x = determinant * (r11 * b.x - r01 * b.y),
		y = determinant * (r00 * b.y - r10 * b.x),
	)
}

/**
 * Solve A * x = b, where b is a column vector. This is more efficient
 * than computing the inverse in one-shot cases.
 */
internal fun IMatrix2d<*>.solve(b: Vector2d): Vector2d {
	var determinant = determinant!!

	if (determinant != 0.0) {
		determinant = 1.0 / determinant
	}

	return Vector2d(
		x = determinant * (r11 * b.x - r01 * b.y),
		y = determinant * (r00 * b.y - r10 * b.x),
	)
}