package ru.dbotthepony.kbox2d.api

import ru.dbotthepony.kbox2d.dynamics.B2Fixture
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import ru.dbotthepony.kbox2d.dynamics.joint.AbstractJoint
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

/**
 * Implement this interface to provide collision filtering. In other words, you can implement
 * this class if you want finer control over contact creation.
 */
interface IContactFilter {
	/**
	 * Return true if contact calculations should be performed between these two shapes.
	 * @warning for performance reasons this is only called when the AABBs begin to overlap.
	 */
	fun shouldCollide(fixtureA: B2Fixture, fixtureB: B2Fixture): Boolean
}

/**
 * Joints and fixtures are destroyed when their associated
 * body is destroyed. Implement this listener so that you
 * may nullify references to these joints and shapes.
 */
interface IDestructionListener {
	/**
	 * Called when any joint is about to be destroyed due
	 * to the destruction of one of its attached bodies.
	 */
	fun sayGoodbye(joint: AbstractJoint)

	/**
	 * Called when any fixture is about to be destroyed due
	 * to the destruction of its parent body.
	 */
	fun sayGoodbye(fixture: B2Fixture)
}

/**
 * Contact impulses for reporting. Impulses are used instead of forces because
 * sub-step forces may approach infinity for rigid body collisions. These
 * match up one-to-one with the contact points in b2Manifold.
 */
data class ContactImpulse(
	val normalImpulses: DoubleArray,
	val tangentImpulses: DoubleArray,
)

/**
 * Implement this class to get contact information. You can use these results for
 * things like sounds and game logic. You can also get contact results by
 * traversing the contact lists after the time step. However, you might miss
 * some contacts because continuous physics leads to sub-stepping.
 *
 * Additionally you may receive multiple callbacks for the same contact in a
 * single time step.
 *
 * You should strive to make your callbacks efficient because there may be
 * many callbacks per time step.
 * @warning You cannot create/destroy Box2D entities inside these callbacks.
 */
interface IContactListener {
	/**
	 * Called when two fixtures begin to touch.
	 */
	fun beginContact(contact: AbstractContact)

	/**
	 * Called when two fixtures cease to touch.
	 */
	fun endContact(contact: AbstractContact)

	/**
	 * This is called after a contact is updated. This allows you to inspect a
	 * contact before it goes to the solver. If you are careful, you can modify the
	 * contact manifold (e.g. disable contact).
	 *
	 * A copy of the old manifold is provided so that you can detect changes.
	 *
	 * Note: this is called only for awake bodies.
	 *
	 * Note: this is called even when the number of contact points is zero.
	 *
	 * Note: this is not called for sensors.
	 *
	 * Note: if you set the number of contact points to zero, you will not
	 * get an EndContact callback. However, you may get a BeginContact callback
	 * the next step.
	 */
	fun preSolve(contact: AbstractContact, oldManifold: Manifold)

	/**
	 * This lets you inspect a contact after the solver is finished. This is useful
	 * for inspecting impulses.
	 *
	 * Note: the contact manifold does not include time of impact impulses, which can be
	 * arbitrarily large if the sub-step is small. Hence the impulse is provided explicitly
	 * in a separate data structure.
	 *
	 * Note: this is only called for contacts that are touching, solid, and awake.
	 */
	fun postSolve(contact: AbstractContact, impulse: ContactImpulse)
}

/**
 * Callback class for AABB queries.
 * See IWorld#query
 */
fun interface IQueryCallback {
	/**
	 * Called for each fixture found in the query AABB.
	 * @return false to terminate the query.
	 */
	fun reportFixture(fixture: B2Fixture): Boolean
}

/**
 * Callback class for ray casts.
 * See IWorld#rayCast
 */
fun interface IRayCastCallback {
	/**
	 * Called for each fixture found in the query. You control how the ray cast
	 * proceeds by returning a float:
	 *
	 * return -1: ignore this fixture and continue
	 *
	 * return 0: terminate the ray cast
	 *
	 * return fraction: clip the ray to this point
	 *
	 * return 1: don't clip the ray and continue
	 *
	 * @param fixture the fixture hit by the ray
	 * @param point the point of initial intersection
	 * @param normal the normal vector at the point of intersection
	 * @param fraction the fraction along the ray at the point of intersection
	 * @return -1 to filter, 0 to terminate, fraction to clip the ray for
	 * closest hit, 1 to continue
	 */
	fun reportFixture(fixture: B2Fixture, point: Vector2d, normal: Vector2d, fraction: Double): Double
}
