package ru.dbotthepony.kbox2d.collision

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.collision.shapes.ChainShape
import ru.dbotthepony.kbox2d.collision.shapes.CircleShape
import ru.dbotthepony.kbox2d.collision.shapes.EdgeShape
import ru.dbotthepony.kbox2d.collision.shapes.PolygonShape
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times
import java.util.*

var b2_gjkCalls = 0
	private set

var b2_gjkIters = 0
	private set

var b2_gjkMaxIters = 0
	private set

/**
 * A distance proxy is used by the GJK algorithm.
 * It encapsulates any shape.
 */
class DistanceProxy(shape: IShape<*>, index: Int) {
	val vertices: List<Vector2d>
	val radius: Double

	init {
		when (shape.type) {
			IShape.Type.CIRCLE -> {
				val circle = shape as CircleShape
				vertices = listOf(circle.position)
				radius = circle.radius
			}

			IShape.Type.EDGE -> {
				val edge = shape as EdgeShape
				vertices = listOf(edge.vertex1, edge.vertex2)
				radius = edge.radius
			}

			IShape.Type.POLYGON -> {
				val polygon = shape as PolygonShape
				vertices = Collections.unmodifiableList(polygon.vertices)
				radius = polygon.radius
			}

			IShape.Type.CHAIN -> {
				val chain = shape as ChainShape
				val faceA = chain.vertices[index]
				val faceB: Vector2d

				if (index + 1 < chain.vertices.size) {
					faceB = chain.vertices[index + 1]
				} else {
					faceB = chain.vertices[0]
				}

				vertices = listOf(faceA, faceB)
				radius = chain.radius
			}
		}
	}

	/**
	 * Get the supporting vertex index in the given direction.
	 */
	fun getSupport(d: Vector2d): Int {
		var bestIndex = 0
		var bestValue = b2Dot(vertices[0], d)

		for (i in 1 until vertices.size) {
			val value = b2Dot(vertices[i], d)

			if (value > bestValue) {
				bestIndex = i
				bestValue = value
			}
		}

		return bestIndex
	}

	/**
	 * Get the supporting vertex in the given direction.
	 */
	fun getSupportVertex(d: Vector2d): Vector2d {
		return vertices[getSupport(d)]
	}
}

data class SimplexVertex(
	var wA: Vector2d = Vector2d.ZERO,		// support point in proxyA
	var wB: Vector2d = Vector2d.ZERO,		// support point in proxyB
	var w: Vector2d = Vector2d.ZERO,		// wB - wA
	var a: Double = 0.0,			// barycentric coordinate for closest point
	var indexA: Int = 0,		// wA index
	var indexB: Int = 0,		// wB index
) {
	fun load(other: SimplexVertex) {
		wA = other.wA
		wB = other.wB
		w = other.w
		a = other.a
		indexA = other.indexA
		indexB = other.indexB
	}
}

data class WitnessPoints(
	val pA: Vector2d,
	val pB: Vector2d,
)

class Simplex() {
	var count: Int = 0
	val v1 = SimplexVertex()
	val v2 = SimplexVertex()
	val v3 = SimplexVertex()
	val vertices = arrayListOf(v1, v2, v3)

	constructor(
		cache: SimplexCache,
		proxyA: DistanceProxy,
		transformA: Transform,
		proxyB: DistanceProxy,
		transformB: Transform,
	) : this() {
		check(cache.count <= 3)
		count = cache.count

		// Copy data from cache.
		for (i in 0 until count) {
			val v = vertices[i]
			v.indexA = cache.indexA[i]
			v.indexB = cache.indexB[i]
			val wALocal = proxyA.vertices[v.indexA]
			val wBLocal = proxyB.vertices[v.indexB]
			v.wA = transformA.times(wALocal)
			v.wB = transformB.times(wBLocal)
			v.w = v.wB - v.wA
		}

		// Compute the new simplex metric, if it is substantially different than
		// old metric then flush the simplex.
		if (count > 1) {
			val metric1 = cache.metric
			val metric2 = getMetric()

			if (metric2 < 0.5 * metric1 || 2.0 * metric1 < metric2 || metric2 < b2_epsilon) {
				// Reset the simplex.
				count = 0
			}
		}

		// If the cache is empty or invalid ...
		if (count == 0) {
			val v = v1
			v.indexA = 0
			v.indexB = 0
			val wALocal = proxyA.vertices[0]
			val wBLocal = proxyB.vertices[0]
			v.wA = transformA.times(wALocal)
			v.wB = transformB.times(wBLocal)
			v.w = v.wB - v.wA;
			v.a = 1.0
			count = 1
		}
	}

	fun writeCache(): SimplexCache {
		return SimplexCache(
			metric = getMetric(),
			count = count,

			indexA = IntArray(count).also {
				for (i in 0 until count) {
					it[i] = vertices[i].indexA
				}
			},

			indexB = IntArray(count).also {
				for (i in 0 until count) {
					it[i] = vertices[i].indexB
				}
			}
		)
	}

	fun getSearchDirection(): Vector2d {
		return when (count) {
			1 -> -v1.w
			2 -> {
				val e12 = v2.w - v1.w
				val sgn = b2Cross(e12, -v1.w)

				if (sgn > 0.0f) {
					// Origin is left of e12.
					return b2Cross(1.0, e12)
				} else {
					// Origin is right of e12.
					return b2Cross(e12, 1.0)
				}
			}
			else -> throw IllegalStateException(count.toString())
		}
	}

	fun getClosestPoint(): Vector2d {
		return when (count) {
			1 -> v1.w
			2 -> v1.a * v1.w + v2.a * v2.w
			3 -> Vector2d.ZERO
			else -> throw IllegalStateException(count.toString())
		}
	}

	fun getWitnessPoints(): WitnessPoints {
		return when (count) {
			1 -> WitnessPoints(v1.wA, v1.wB)
			2 -> WitnessPoints(
				pA = v1.a * v1.wA + v2.a * v2.wA,
				pB = v1.a * v1.wB + v2.a * v2.wB,
			)
			3 -> {
				val pA = v1.a * v1.wA + v2.a * v2.wA + v3.a * v3.wA

				WitnessPoints(
					pA = pA,
					pB = pA
				)
			}
			else -> throw IllegalStateException(count.toString())
		}
	}

	fun getMetric(): Double {
		return when (count) {
			1 -> 0.0
			2 -> b2Distance(v1.w, v2.w)
			3 -> b2Cross(v2.w - v1.w, v3.w - v1.w)
			else -> throw IllegalStateException(count.toString())
		}
	}

	/**
	 * Solve a line segment using barycentric coordinates.
	 *
	 * p = a1 * w1 + a2 * w2
	 * a1 + a2 = 1
	 *
	 * The vector from the origin to the closest point on the line is
	 * perpendicular to the line.
	 * e12 = w2 - w1
	 * dot(p, e) = 0
	 * a1 * dot(w1, e) + a2 * dot(w2, e) = 0
	 *
	 * 2-by-2 linear system
	 * [1      1     ][a1] = [1]
	 * [w1.e12 w2.e12][a2] = [0]
	 *
	 * Define
	 * d12_1 =  dot(w2, e12)
	 * d12_2 = -dot(w1, e12)
	 * d12 = d12_1 + d12_2
	 *
	 * Solution
	 * a1 = d12_1 / d12
	 * a2 = d12_2 / d12
	 */
	fun solve2() {
		val w1 = v1.w
		val w2 = v2.w
		val e12 = w2 - w1

		// w1 region
		val d12_2 = b2Dot(-w1, e12)
		if (d12_2 <= 0.0) {
			// a2 <= 0, so we clamp it to 0
			v1.a = 1.0
			count = 1
			return
		}

		// w2 region
		val d12_1 = b2Dot(w2, e12)
		if (d12_1 <= 0.0) {
			// a1 <= 0, so we clamp it to 0
			v2.a = 1.0
			count = 1
			v1.load(v2) // TODO
			return
		}

		// Must be in e12 region.
		val inv_d12 = 1.0f / (d12_1 + d12_2)
		v1.a = d12_1 * inv_d12
		v2.a = d12_2 * inv_d12
		count = 2
	}

	/**
	 * Possible regions:
	 * - points[2]
	 * - edge points[0]-points[2]
	 * - edge points[1]-points[2]
	 * - inside the triangle
	 */
	fun solve3() {
		val w1 = v1.w
		val w2 = v2.w
		val w3 = v3.w

		// Edge12
		// [1      1     ][a1] = [1]
		// [w1.e12 w2.e12][a2] = [0]
		// a3 = 0
		val e12 = w2 - w1
		val w1e12 = b2Dot(w1, e12)
		val w2e12 = b2Dot(w2, e12)
		val d12_1 = w2e12
		val d12_2 = -w1e12

		// Edge13
		// [1      1     ][a1] = [1]
		// [w1.e13 w3.e13][a3] = [0]
		// a2 = 0
		val e13 = w3 - w1
		val w1e13 = b2Dot(w1, e13)
		val w3e13 = b2Dot(w3, e13)
		val d13_1 = w3e13
		val d13_2 = -w1e13

		// Edge23
		// [1      1     ][a2] = [1]
		// [w2.e23 w3.e23][a3] = [0]
		// a1 = 0
		val e23 = w3 - w2
		val w2e23 = b2Dot(w2, e23)
		val w3e23 = b2Dot(w3, e23)
		val d23_1 = w3e23
		val d23_2 = -w2e23

		// Triangle123
		val n123 = b2Cross(e12, e13)

		val d123_1 = n123 * b2Cross(w2, w3)
		val d123_2 = n123 * b2Cross(w3, w1)
		val d123_3 = n123 * b2Cross(w1, w2)

		// w1 region
		if (d12_2 <= 0.0 && d13_2 <= 0.0) {
			v1.a = 1.0
			count = 1
			return
		}

		// e12
		if (d12_1 > 0.0 && d12_2 > 0.0 && d123_3 <= 0.0) {
			val inv_d12 = 1.0 / (d12_1 + d12_2)
			v1.a = d12_1 * inv_d12
			v2.a = d12_2 * inv_d12
			count = 2
			return
		}

		// e13
		if (d13_1 > 0.0 && d13_2 > 0.0 && d123_2 <= 0.0) {
			val inv_d13 = 1.0 / (d13_1 + d13_2)
			v1.a = d13_1 * inv_d13
			v3.a = d13_2 * inv_d13
			count = 2
			v2.load(v3) // TODO
			return
		}

		// w2 region
		if (d12_1 <= 0.0 && d23_2 <= 0.0) {
			v2.a = 1.0
			count = 1
			v1.load(v2) // TODO
			return
		}

		// w3 region
		if (d13_1 <= 0.0 && d23_1 <= 0.0) {
			v3.a = 1.0
			count = 1
			v1.load(v3)
			return
		}

		// e23
		if (d23_1 > 0.0 && d23_2 > 0.0 && d123_1 <= 0.0) {
			val inv_d23 = 1.0 / (d23_1 + d23_2)
			v2.a = d23_1 * inv_d23
			v3.a = d23_2 * inv_d23
			count = 2
			v1.load(v3)
			return
		}

		// Must be in triangle123
		val inv_d123 = 1.0 / (d123_1 + d123_2 + d123_3)
		v1.a = d123_1 * inv_d123
		v2.a = d123_2 * inv_d123
		v3.a = d123_3 * inv_d123
		count = 3
	}
}

private const val k_maxIters = 20

/**
 * Compute the closest points between two shapes. Supports any combination of:
 * b2CircleShape, b2PolygonShape, b2EdgeShape. The simplex cache is input/output.
 * On the first call set b2SimplexCache.count to zero.
 */
fun b2Distance(
	cache: SimplexCache,
	proxyA: DistanceProxy,
	proxyB: DistanceProxy,
	transformA: Transform,
	transformB: Transform,
	useRadii: Boolean = false
): DistanceOutput {
	// Initialize the simplex.
	val simplex = Simplex(cache, proxyA, transformA, proxyB, transformB)

	// Get simplex vertices as an array.
	val vertices = simplex.vertices

	// These store the vertices of the last simplex so that we
	// can check for duplicates and prevent cycling.
	val saveA = IntArray(3)
	val saveB = IntArray(3)
	var saveCount: Int

	var iter = 0
	while (iter < k_maxIters) {
		// Copy simplex so we can identify duplicates.
		saveCount = simplex.count

		for (i in 0 until saveCount) {
			saveA[i] = vertices[i].indexA
			saveB[i] = vertices[i].indexB
		}

		when (simplex.count) {
			1 -> {}
			2 -> simplex.solve2()
			3 -> simplex.solve3()
			else -> throw IllegalStateException(simplex.count.toString())
		}

		if (simplex.count == 3) {
			// If we have 3 points, then the origin is in the corresponding triangle.
			break
		}

		// Get search direction.
		val d = simplex.getSearchDirection()

		// Ensure the search direction is numerically fit.
		if (d.lengthSquared < b2_epsilon * b2_epsilon) {
			// The origin is probably contained by a line segment
			// or triangle. Thus the shapes are overlapped.

			// We can't return zero here even though there may be overlap.
			// In case the simplex is a point, segment, or triangle it is difficult
			// to determine if the origin is contained in the CSO or very close to it.
			break
		}

		// Compute a tentative new simplex vertex using support points.
		val vertex = vertices[simplex.count]
		vertex.indexA = proxyA.getSupport(transformA.q.timesT(-d))
		vertex.wA = transformA.times(proxyA.vertices[vertex.indexA])
		vertex.indexB = proxyB.getSupport(transformB.q.timesT(d))
		vertex.wB = transformB.times(proxyB.vertices[vertex.indexB])
		vertex.w = vertex.wB - vertex.wA

		// Iteration count is equated to the number of support point calls.
		iter++
		b2_gjkIters++

		// Check for duplicate support points. This is the main termination criteria.
		var duplicate = false

		for (i in 0 until saveCount) {
			if (vertex.indexA == saveA[i] && vertex.indexB == saveB[i]) {
				duplicate = true
				break
			}
		}

		// If we found a duplicate support point we must exit to avoid cycling.
		if (duplicate) {
			break
		}

		// New vertex is ok and needed.
		simplex.count++
	}

	b2_gjkMaxIters = b2_gjkMaxIters.coerceAtLeast(iter)

	// Prepare output.
	var (pointA, pointB) = simplex.getWitnessPoints()
	// Cache the simplex.
	val newCache = simplex.writeCache()

	var distance = pointA.distance(pointB)

	// Apply radii if requested
	if (useRadii) {
		if (distance < b2_epsilon) {
			// Shapes are too close to safely compute normal
			val p = 0.5 * (pointA + pointB)
			pointA = p
			pointB = p
			distance = 0.0
		} else {
			// Keep closest points on perimeter even if overlapped, this way
			// the points move smoothly.
			val rA = proxyA.radius;
			val rB = proxyB.radius;
			val normal = (pointB - pointA).normalized;
			distance = (distance - rA - rB).coerceAtLeast(0.0)
			pointA += rA * normal
			pointB -= rB * normal
		}
	}

	return DistanceOutput(
		pointA = pointA,
		pointB = pointB,
		distance = distance,
		newCache = newCache,
		iterations = iter,
	)
}

fun b2Distance(
	cache: SimplexCache,
	input: DistanceInput
): DistanceOutput {
	val (proxyA, proxyB, transformA, transformB, useRadii) = input
	return b2Distance(cache, proxyA, proxyB, transformA, transformB, useRadii)
}

private const val gjk_tolerance = 0.5 * b2_linearSlop

/**
 * GJK-raycast
 * Algorithm by Gino van den Bergen.
 * "Smooth Mesh Contacts with GJK" in Game Physics Pearls. 2010
 */
fun b2ShapeCast(
	output: ShapeCastOutput,
	proxyA: DistanceProxy,
	proxyB: DistanceProxy,
	xfA: Transform,
	xfB: Transform,
	r: Vector2d,
): Boolean {
	output.iterations = 0
	output.lambda = 1.0
	output.normal = Vector2d.ZERO
	output.point = Vector2d.ZERO

	val radiusA = proxyA.radius.coerceAtLeast(b2_polygonRadius)
	val radiusB = proxyB.radius.coerceAtLeast(b2_polygonRadius)

	val radius = radiusA + radiusB
	var n = Vector2d.ZERO
	var lambda = 0.0

	// Initial simplex
	val simplex = Simplex()

	// Get simplex vertices as an array.
	val vertices = simplex.vertices

	// Get support point in -r direction
	var indexA = proxyA.getSupport(xfA.q.timesT(-r))
	var wA = xfA.times(proxyA.vertices[indexA])

	var indexB = proxyB.getSupport(xfB.q.timesT(r))
	var wB = xfB.times(proxyB.vertices[indexB])
	var v = wA - wB

	// Sigma is the target distance between polygons
	val sigma = b2_polygonRadius.coerceAtLeast(radius - b2_polygonRadius)

	// Main iteration loop.
	var iter = 0

	while (iter < k_maxIters && v.length - sigma > gjk_tolerance) {
		check(simplex.count < 3) { simplex.count }

		output.iterations++

		// Support in direction -v (A - B)
		indexA = proxyA.getSupport(xfA.q.timesT(-v))
		wA = xfA.times(proxyA.vertices[indexA])

		indexB = proxyB.getSupport(xfB.q.timesT(v))
		wB = xfB.times(proxyB.vertices[indexB])

		val p = wA - wB

		// -v is a normal at p
		v = v.normalized

		// Intersect ray with plane
		val vp = b2Dot(v, p)
		val vr = b2Dot(v, r)

		if (vp - sigma > lambda * vr) {
			if (vr <= 0.0) {
				return false
			}

			lambda = (vp - sigma) / vr

			if (lambda >= 1.0) {
				return false
			}

			n = -v
			simplex.count = 0
		}

		// Reverse simplex since it works with B - A.
		// Shift by lambda * r because we want the closest point to the current clip point.
		// Note that the support point p is not shifted because we want the plane equation
		// to be formed in unshifted space.
		val vertex = vertices[simplex.count]
		vertex.indexA = indexB
		vertex.wA = wB + lambda * r
		vertex.indexB = indexA
		vertex.wB = wA
		vertex.w = vertex.wB - vertex.wA
		vertex.a = 1.0
		simplex.count++

		when (simplex.count) {
			1 -> {}
			2 -> simplex.solve2()
			3 -> simplex.solve3()
			else -> throw IllegalStateException(simplex.count.toString())
		}

		// If we have 3 points, then the origin is in the corresponding triangle.
		if (simplex.count == 3) {
			// Overlap
			return false
		}

		// Get search direction.
		v = simplex.getClosestPoint()

		// Iteration count is equated to the number of support point calls.
		iter++
	}

	if (iter == 0) {
		// Initial overlap
		return false
	}

	// Prepare output.
	val (_, pointA) = simplex.getWitnessPoints()

	if (v.lengthSquared > 0.0) {
		n = (-v).normalized
	}

	output.point = pointA + radiusA * n
	output.normal = n
	output.lambda = lambda
	output.iterations = iter

	return true
}

fun b2ShapeCast(output: ShapeCastOutput, input: ShapeCastInput): Boolean {
	val (proxyA, proxyB, transformA, transformB, translationB) = input
	return b2ShapeCast(output, proxyA, proxyB, transformA, transformB, translationB)
}
