package ru.dbotthepony.kbox2d.dynamics.internal

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.collision.WorldManifold
import ru.dbotthepony.kbox2d.dynamics.contact.AbstractContact
import ru.dbotthepony.kbox2d.dynamics.B2Body
import ru.dbotthepony.kvector.matrix.ndouble.Matrix2d
import ru.dbotthepony.kvector.matrix.ndouble.MutableMatrix2d
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times
import java.lang.IllegalArgumentException
import kotlin.math.absoluteValue

private const val g_blockSolve = true

// Ensure a reasonable condition number.
private const val k_maxConditionNumber = 1000.0

private const val k_errorTol = 1E-3

private const val B2_DEBUG_SOLVER = false

internal data class VelocityCostantPoint(
	var rA: Vector2d = Vector2d.ZERO,
	var rB: Vector2d = Vector2d.ZERO,
	var normalImpulse: Double = 0.0,
	var tangentImpulse: Double = 0.0,
	var normalMass: Double = 0.0,
	var tangentMass: Double = 0.0,
	var velocityBias: Double = 0.0,
)

internal data class ContactVelocityConstraint(
	var points: Array<VelocityCostantPoint>,
	var normal: Vector2d,
	var normalMass: MutableMatrix2d,
	var K: MutableMatrix2d,
	var indexA: Int,
	var indexB: Int,
	var invMassA: Double,
	var invMassB: Double,
	var invIA: Double,
	var invIB: Double,
	var friction: Double,
	var restitution: Double,
	var threshold: Double,
	var tangentSpeed: Double,
	var contactIndex: Int,
)

internal data class ContactSolverDef(
	val step: B2TimeStep,
	val contacts: List<AbstractContact>,
	val positions: List<B2Position>,
	val velocities: List<B2Velocity>,
)

internal data class ContactPositionConstraint(
	val localPoints: Array<Vector2d>,
	var localNormal: Vector2d,
	var localPoint: Vector2d,
	var indexA: Int,
	var indexB: Int,
	var invMassA: Double,
	var invMassB: Double,
	var localCenterA: Vector2d,
	var localCenterB: Vector2d,
	var invIA: Double,
	var invIB: Double,
	var radiusA: Double,
	var radiusB: Double,
	val type: Manifold.Type?
)

internal class PositionSolverManifold(pc: ContactPositionConstraint, xfA: Transform, xfB: Transform, index: Int) {
	val normal: Vector2d
	val point: Vector2d
	val separation: Double

	operator fun component1() = normal
	operator fun component2() = point
	operator fun component3() = separation

	init {
		check(pc.localPoints.isNotEmpty()) { "localPoints is empty" }

		when (pc.type) {
			Manifold.Type.CIRCLES -> {
				val pointA = b2Mul(xfA, pc.localPoint)
				val pointB = b2Mul(xfB, pc.localPoints[0])
				normal = (pointB - pointA).normalized
				point = 0.5 * (pointA + pointB)
				separation = b2Dot(pointB - pointA, normal) - pc.radiusA - pc.radiusB
			}

			Manifold.Type.FACE_A -> {
				normal = b2Mul(xfA.q, pc.localNormal)
				val planePoint = b2Mul(xfA, pc.localPoint)

				val clipPoint = b2Mul(xfB, pc.localPoints[index])
				separation = b2Dot(clipPoint - planePoint, normal) - pc.radiusA - pc.radiusB
				point = clipPoint
			}

			Manifold.Type.FACE_B -> {
				val normal = b2Mul(xfB.q, pc.localNormal)
				val planePoint = b2Mul(xfB, pc.localPoint)

				val clipPoint = b2Mul(xfA, pc.localPoints[index])
				separation = b2Dot(clipPoint - planePoint, normal) - pc.radiusA - pc.radiusB
				point = clipPoint

				// Ensure normal points from A to B
				this.normal = -normal
			}

			else -> throw IllegalArgumentException("Position Constraint $pc has manifold of null (unknown) type")
		}
	}
}

internal class ContactSolver(
	val step: B2TimeStep,
	val positions: List<B2Position>,
	val velocities: List<B2Velocity>,
	val contacts: List<AbstractContact>,
) {
	constructor(def: ContactSolverDef) : this(
		def.step,
		def.positions,
		def.velocities,
		def.contacts,
	)

	val positionConstraints = ArrayList<ContactPositionConstraint>()
	val velocityConstraints = ArrayList<ContactVelocityConstraint>()

	init {
		// Initialize position independent portions of the constraints.
		for ((i, contact) in contacts.withIndex()) {
			val fixtureA = contact.fixtureA
			val fixtureB = contact.fixtureB
			val shapeA = fixtureA.shape
			val shapeB = fixtureB.shape
			val radiusA = shapeA.radius
			val radiusB = shapeB.radius
			val bodyA = fixtureA.body as B2Body
			val bodyB = fixtureB.body as B2Body
			val manifold = contact.manifold

			check(manifold.points.isNotEmpty()) { "Manifold points at $i are empty" }

			val vc = ContactVelocityConstraint(
				friction = contact.friction,
				restitution = contact.restitution,
				threshold = contact.restitutionThreshold,
				tangentSpeed = contact.tangentSpeed,
				indexA = bodyA.islandIndex,
				indexB = bodyB.islandIndex,
				invMassA = bodyA.invMass,
				invMassB = bodyB.invMass,
				invIA = bodyA.rotInertiaInv,
				invIB = bodyB.rotInertiaInv,
				contactIndex = i,
				K = MutableMatrix2d.zero(),
				normalMass = MutableMatrix2d.zero(),
				normal = Vector2d.ZERO,
				points = Array(manifold.points.size) { VelocityCostantPoint() }
			)

			velocityConstraints.add(vc)

			val pc = ContactPositionConstraint(
				indexA = bodyA.islandIndex,
				indexB = bodyB.islandIndex,
				invMassA = bodyA.invMass,
				invMassB = bodyB.invMass,
				localCenterA = bodyA.sweep.localCenter,
				localCenterB = bodyB.sweep.localCenter,
				invIA = bodyA.rotInertiaInv,
				invIB = bodyB.rotInertiaInv,
				localNormal = manifold.localNormal,
				localPoint = manifold.localPoint,
				localPoints = Array(manifold.points.size) { Vector2d.ZERO },
				radiusA = radiusA,
				radiusB = radiusB,
				type = manifold.type,
			)

			positionConstraints.add(pc)

			for (j in manifold.points.indices) {
				val cp = manifold.points[j]
				val vcp = vc.points[j]

				if (step.warmStarting) {
					vcp.normalImpulse = cp.normalImpulse * step.dtRatio
					vcp.tangentImpulse = cp.tangentImpulse * step.dtRatio
				}

				pc.localPoints[j] = cp.localPoint
			}
		}
	}

	/**
	 * Initialize position dependent portions of the velocity constraints.
	 */
	fun initializeVelocityConstraints() {
		for (i in contacts.indices) {
			val vc = velocityConstraints[i]
			val pc = positionConstraints[i]

			val radiusA = pc.radiusA
			val radiusB = pc.radiusB
			val manifold = contacts[vc.contactIndex].manifold

			val indexA = vc.indexA
			val indexB = vc.indexB

			val mA = vc.invMassA
			val mB = vc.invMassB
			val iA = vc.invIA
			val iB = vc.invIB

			val localCenterA = pc.localCenterA
			val localCenterB = pc.localCenterB

			val cA = positions[indexA].c
			val aA = positions[indexA].a
			val vA = velocities[indexA].v
			val wA = velocities[indexA].w

			val cB = positions[indexB].c
			val aB = positions[indexB].a
			val vB = velocities[indexB].v
			val wB = velocities[indexB].w

			check(manifold.points.isNotEmpty()) { "Manifold at $i is empty" }

			val xfA = Transform()
			val xfB = Transform()

			xfA.rotation.set(aA)
			xfB.rotation.set(aB)

			xfA.position = cA - b2Mul(xfA.q, localCenterA)
			xfB.position = cB - b2Mul(xfB.q, localCenterB)

			val worldManifold = WorldManifold(manifold, xfA, radiusA, xfB, radiusB)

			vc.normal = worldManifold.normal

			for ((j, vcp) in vc.points.withIndex()) {
				vcp.rA = worldManifold.points[j] - cA
				vcp.rB = worldManifold.points[j] - cB

				val rnA = b2Cross(vcp.rA, vc.normal)
				val rnB = b2Cross(vcp.rB, vc.normal)

				val kNormal = mA + mB + iA * rnA * rnA + iB * rnB * rnB

				vcp.normalMass = if (kNormal > 0.0) 1.0 / kNormal else 0.0

				val tangent = b2Cross(vc.normal, 1.0)

				val rtA = b2Cross(vcp.rA, tangent)
				val rtB = b2Cross(vcp.rB, tangent)

				val kTangent = mA + mB + iA * rtA * rtA + iB * rtB * rtB

				vcp.tangentMass = if (kTangent > 0.0) 1.0 / kTangent else 0.0

				// Setup a velocity bias for restitution.
				// vcp.velocityBias = 0.0
				val vRel = b2Dot(vc.normal, vB + b2Cross(wB, vcp.rB) - vA - b2Cross(wA, vcp.rA))

				if (vRel < -vc.threshold) {
					vcp.velocityBias = -vc.restitution * vRel
				}
			}

			// If we have two points, then prepare the block solver.
			if (vc.points.size == 2 && g_blockSolve) {
				val vcp1 = vc.points[0]
				val vcp2 = vc.points[1]

				val rn1A = b2Cross(vcp1.rA, vc.normal)
				val rn1B = b2Cross(vcp1.rB, vc.normal)
				val rn2A = b2Cross(vcp2.rA, vc.normal)
				val rn2B = b2Cross(vcp2.rB, vc.normal)

				val k11 = mA + mB + iA * rn1A * rn1A + iB * rn1B * rn1B
				val k22 = mA + mB + iA * rn2A * rn2A + iB * rn2B * rn2B
				val k12 = mA + mB + iA * rn1A * rn2A + iB * rn1B * rn2B

				if (k11 * k11 < k_maxConditionNumber * (k11 * k22 - k12 * k12)) {
					// K is safe to invert.
					// k11 k12
					// k12 k22
					//
					// vc->K.ex.Set(k11, k12);
					// vc->K.ey.Set(k12, k22);

					vc.K.r00 = k11
					vc.K.r10 = k12
					vc.K.r01 = k12
					vc.K.r11 = k22
					vc.normalMass = (vc.K.inverse ?: Matrix2d.ZERO).toMutableMatrix2d()
				} else {
					// The constraints are redundant, just use one.
					// TODO_ERIN use deepest?
					vc.points = Array(1) { vc.points[it] }
				}
			}
		}
	}

	/**
	 * Warm start.
	 */
	fun warmStart() {
		for (vc in velocityConstraints) {
			val indexA = vc.indexA
			val indexB = vc.indexB
			val mA = vc.invMassA
			val iA = vc.invIA
			val mB = vc.invMassB
			val iB = vc.invIB

			var vA = velocities[indexA].v
			var wA = velocities[indexA].w
			var vB = velocities[indexB].v
			var wB = velocities[indexB].w

			val normal = vc.normal
			val tangent = b2Cross(normal, 1.0)

			for (vcp in vc.points) {
				val P = normal * vcp.normalImpulse + tangent * vcp.tangentImpulse
				wA -= iA * b2Cross(vcp.rA, P)
				vA -= P * mA
				wB += iB * b2Cross(vcp.rB, P)
				vB += P * mB
			}

			velocities[indexA].v = vA
			velocities[indexA].w = wA
			velocities[indexB].v = vB
			velocities[indexB].w = wB
		}
	}

	fun solveVelocityConstraints() {
		for (vc in velocityConstraints) {
			val indexA = vc.indexA
			val indexB = vc.indexB
			val mA = vc.invMassA
			val iA = vc.invIA
			val mB = vc.invMassB
			val iB = vc.invIB

			var vA = velocities[indexA].v
			var wA = velocities[indexA].w
			var vB = velocities[indexB].v
			var wB = velocities[indexB].w

			val normal = vc.normal
			val tangent = b2Cross(normal, 1.0)
			val friction = vc.friction

			check(vc.points.size == 1 || vc.points.size == 2) { "Unexpected points amount: ${vc.points.size}" }

			// Solve tangent constraints first because non-penetration is more important
			// than friction.
			for (vcp in vc.points) {
				// Relative velocity at contact
				val dv = vB + vB + b2Cross(wB, vcp.rB) - vA - b2Cross(wA, vcp.rA)

				// Compute tangent force
				val vt = b2Dot(dv, tangent) - vc.tangentSpeed
				var lambda = vcp.tangentMass * (-vt)

				// b2Clamp the accumulated force
				val maxFriction = friction * vcp.normalImpulse
				val newImpulse = b2Clamp(vcp.tangentImpulse + lambda, -maxFriction, maxFriction)

				lambda = newImpulse - vcp.tangentImpulse
				vcp.tangentImpulse = newImpulse

				// Apply contact impulse
				val P = lambda * tangent

				vA -= mA * P
				wA -= iA * b2Cross(vcp.rA, P)

				vB += mB * P
				wB += iB * b2Cross(vcp.rB, P)
			}

			// Solve normal constraints
			if (vc.points.size == 1 || !g_blockSolve) {
				for (vcp in vc.points) {
					// Relative velocity at contact
					val dv = vB + b2Cross(wB, vcp.rB) - vA - b2Cross(wA, vcp.rA)

					// Compute normal impulse
					val vn = b2Dot(dv, normal)
					var lambda = -vcp.normalMass * (vn - vcp.velocityBias)

					// b2Clamp the accumulated impulse
					val newImpulse = b2Max(vcp.normalImpulse + lambda, 0.0)
					lambda = newImpulse - vcp.normalImpulse
					vcp.normalImpulse = newImpulse

					// Apply contact impulse
					val P = lambda * normal
					vA -= mA * P
					wA -= iA * b2Cross(vcp.rA, P)

					vB += mB * P
					wB += iB * b2Cross(vcp.rB, P)
				}
			} else {
				// Block solver developed in collaboration with Dirk Gregorius (back in 01/07 on Box2D_Lite).
				// Build the mini LCP for this contact patch
				//
				// vn = A * x + b, vn >= 0, x >= 0 and vn_i * x_i = 0 with i = 1..2
				//
				// A = J * W * JT and J = ( -n, -r1 x n, n, r2 x n )
				// b = vn0 - velocityBias
				//
				// The system is solved using the "Total enumeration method" (s. Murty). The complementary constraint vn_i * x_i
				// implies that we must have in any solution either vn_i = 0 or x_i = 0. So for the 2D contact problem the cases
				// vn1 = 0 and vn2 = 0, x1 = 0 and x2 = 0, x1 = 0 and vn2 = 0, x2 = 0 and vn1 = 0 need to be tested. The first valid
				// solution that satisfies the problem is chosen.
				//
				// In order to account of the accumulated impulse 'a' (because of the iterative nature of the solver which only requires
				// that the accumulated impulse is clamped and not the incremental impulse) we change the impulse variable (x_i).
				//
				// Substitute:
				//
				// x = a + d
				//
				// a := old total impulse
				// x := new total impulse
				// d := incremental impulse
				//
				// For the current iteration we extend the formula for the incremental impulse
				// to compute the new total impulse:
				//
				// vn = A * d + b
				//    = A * (x - a) + b
				//    = A * x + b - A * a
				//    = A * x + b'
				// b' = b - A * a;

				val cp1 = vc.points[0]
				val cp2 = vc.points[1]

				val a = Vector2d(cp1.normalImpulse, cp2.normalImpulse)

				check(a.x >= 0.0 && a.y >= 0.0) { a }

				// Relative velocity at contact
				var dv1 = vB + b2Cross(wB, cp1.rB) - vA - b2Cross(wA, cp1.rA)
				var dv2 = vB + b2Cross(wB, cp2.rB) - vA - b2Cross(wA, cp2.rA)

				// Compute normal velocity
				var vn1 = b2Dot(dv1, normal)
				var vn2 = b2Dot(dv2, normal)

				var b = Vector2d(
					x = vn1 - cp1.velocityBias,
					y = vn2 - cp2.velocityBias,
				)

				// Compute b'
				b -= b2Mul(vc.K, a)

				// for (;;)
				run {
					//
					// Case 1: vn = 0
					//
					// 0 = A * x + b'
					//
					// Solve for x:
					//
					// x = - inv(A) * b'
					//
					var x = -b2Mul(vc.normalMass, b)

					if (x.x >= 0.0 && x.y >= 0.0) {
						// Get the incremental impulse
						val d = x - a

						// Apply incremental impulse
						val P1 = normal * d.x
						val P2 = normal * d.y

						vA -= mA * (P1 + P2)
						wA -= iA * (b2Cross(cp1.rA, P1) + b2Cross(cp2.rA, P2))

						vB += mB * (P1 + P2)
						wB += iB * (b2Cross(cp1.rB, P1) + b2Cross(cp2.rB, P2))

						// Accumulate
						cp1.normalImpulse = x.x
						cp2.normalImpulse = x.y

						// Postconditions
						if (B2_DEBUG_SOLVER) {
							dv1 = vB + b2Cross(wB, cp1.rB) - vA - b2Cross(wA, cp1.rA)
							dv2 = vB + b2Cross(wB, cp2.rB) - vA - b2Cross(wA, cp2.rA)

							// Compute normal velocity
							vn1 = b2Dot(dv1, normal)
							vn2 = b2Dot(dv2, normal)

							check((vn1 - cp1.velocityBias).absoluteValue < k_errorTol) { (vn1 - cp1.velocityBias).absoluteValue }
							check((vn2 - cp2.velocityBias).absoluteValue < k_errorTol) { (vn2 - cp2.velocityBias).absoluteValue }
						}

						return@run
					}

					//
					// Case 2: vn1 = 0 and x2 = 0
					//
					//   0 = a11 * x1 + a12 * 0 + b1'
					// vn2 = a21 * x1 + a22 * 0 + b2'
					//
					x = Vector2d(x = -cp1.normalMass * b.x)

					vn1 = 0.0
					// vn2 = vc->K.ex.y * x.x + b.y;
					vn2 = vc.K.r10 * x.x + b.y

					if (x.x >= 0.0 && vn2 >= 0.0) {
						// Get the incremental impulse
						val d = x - a

						// Apply incremental impulse
						val P1 = d.x * normal
						val P2 = d.y * normal

						vA -= mA * (P1 + P2)
						wA -= iA * (b2Cross(cp1.rA, P1) + b2Cross(cp2.rA, P2))

						vB += mB * (P1 + P2)
						wB += iB * (b2Cross(cp1.rB, P1) + b2Cross(cp2.rB, P2))

						// Accumulate
						cp1.normalImpulse = x.x
						cp2.normalImpulse = x.y

						if (B2_DEBUG_SOLVER) {
							// Postconditions
							dv1 = vB + b2Cross(wB, cp1.rB) - vA - b2Cross(wA, cp1.rA)

							// Compute normal velocity
							vn1 = b2Dot(dv1, normal)

							check((vn1 - cp1.velocityBias).absoluteValue < k_errorTol) { (vn1 - cp1.velocityBias).absoluteValue }
						}

						return@run
					}

					//
					// Case 3: vn2 = 0 and x1 = 0
					//
					// vn1 = a11 * 0 + a12 * x2 + b1'
					//   0 = a21 * 0 + a22 * x2 + b2'
					//
					x = Vector2d(y = -cp2.normalMass * b.y)
					vn1 = vc.K.r01 * x.y + b.x
					vn2 = 0.0

					if (x.y >= 0.0 && vn1 >= 0.0) {
						// Resubstitute for the incremental impulse
						val d = x - a

						// Apply incremental impulse
						val P1 = d.x * normal
						val P2 = d.y * normal

						vA -= mA * (P1 + P2)
						wA -= iA * (b2Cross(cp1.rA, P1) + b2Cross(cp2.rA, P2))

						vB += mB * (P1 + P2)
						wB += iB * (b2Cross(cp1.rB, P1) + b2Cross(cp2.rB, P2))

						// Accumulate
						cp1.normalImpulse = x.x
						cp2.normalImpulse = x.y

						if (B2_DEBUG_SOLVER) {
							// Postconditions
							dv2 = vB + b2Cross(wB, cp2.rB) - vA - b2Cross(wA, cp2.rA)

							// Compute normal velocity
							vn2 = b2Dot(dv2, normal)

							check((vn2 - cp2.velocityBias).absoluteValue < k_errorTol) { (vn2 - cp2.velocityBias).absoluteValue }
						}

						return@run
					}

					//
					// Case 4: x1 = 0 and x2 = 0
					//
					// vn1 = b1
					// vn2 = b2;
					x = Vector2d.ZERO
					vn1 = b.x
					vn2 = b.y

					if (vn1 >= 0.0 && vn2 >= 0.0) {
						// Resubstitute for the incremental impulse
						val d = x - a

						// Apply incremental impulse
						val P1 = d.x * normal
						val P2 = d.y * normal

						vA -= mA * (P1 + P2)
						wA -= iA * (b2Cross(cp1.rA, P1) + b2Cross(cp2.rA, P2))

						vB += mB * (P1 + P2)
						wB += iB * (b2Cross(cp1.rB, P1) + b2Cross(cp2.rB, P2))

						// Accumulate
						cp1.normalImpulse = x.x
						cp2.normalImpulse = x.y

						return@run
					}

					// No solution, give up. This is hit sometimes, but it doesn't seem to matter.
				}
			}

			velocities[indexA].v = vA
			velocities[indexA].w = wA
			velocities[indexB].v = vB
			velocities[indexB].w = wB
		}
	}

	fun storeImpulses() {
		for (vc in velocityConstraints) {
			val manifold = contacts[vc.contactIndex].manifold

			for ((j, point) in vc.points.withIndex()) {
				manifold.points[j].normalImpulse = point.normalImpulse
				manifold.points[j].tangentImpulse = point.tangentImpulse
			}
		}
	}

	/**
	 * Sequential solver.
 	 */
	fun solvePositionConstraints(): Boolean {
		var minSeparation = 0.0

		for (pc in positionConstraints) {
			val indexA = pc.indexA
			val indexB = pc.indexB
			val localCenterA = pc.localCenterA
			val mA = pc.invMassA
			val iA = pc.invIA
			val localCenterB = pc.localCenterB
			val mB = pc.invMassB
			val iB = pc.invIB

			var cA = positions[indexA].c
			var aA = positions[indexA].a

			var cB = positions[indexB].c
			var aB = positions[indexB].a

			// Solve normal constraints
			for (j in 0 until pc.localPoints.size) {
				val xfA = Transform()
				val xfB = Transform()

				xfA.q.set(aA)
				xfB.q.set(aB)

				xfA.p = cA - b2Mul(xfA.q, localCenterA)
				xfB.p = cB - b2Mul(xfB.q, localCenterB)

				val (normal, point, separation) = PositionSolverManifold(pc, xfA, xfB, j)

				val rA = point - cA
				val rB = point - cB

				// Track max constraint error.
				minSeparation = b2Min(minSeparation, separation)

				// Prevent large corrections and allow slop.
				val C = b2Clamp(b2_baumgarte * (separation + b2_linearSlop), -b2_maxLinearCorrection, 0.0)

				// Compute the effective mass.
				val rnA = b2Cross(rA, normal)
				val rnB = b2Cross(rB, normal)
				val K = mA + mB + iA * rnA * rnA + iB * rnB * rnB

				// Compute normal impulse
				val impulse = if (K > 0.0) -C / K else 0.0

				val P = impulse * normal

				cA -= mA * P
				aA -= iA * b2Cross(rA, P)

				cB += mB * P
				aB += iB * b2Cross(rB, P)
			}

			positions[indexA].c = cA
			positions[indexA].a = aA

			positions[indexB].c = cB
			positions[indexB].a = aB
		}

		// We can't expect minSpeparation >= -b2_linearSlop because we don't
		// push the separation above -b2_linearSlop.
		return minSeparation >= -3.0f * b2_linearSlop
	}

	/**
	 * Sequential position solver for position constraints.
	 */
	fun solveTOIPositionConstraints(toiIndexA: Int, toiIndexB: Int): Boolean {
		var minSeparation = 0.0

		for (pc in positionConstraints) {
			val indexA = pc.indexA
			val indexB = pc.indexB
			val localCenterA = pc.localCenterA
			val localCenterB = pc.localCenterB

			var mA = 0.0
			var iA = 0.0

			if (indexA == toiIndexA || indexA == toiIndexB) {
				mA = pc.invMassA
				iA = pc.invIA
			}

			var mB = 0.0
			var iB = 0.0

			if (indexB == toiIndexA || indexB == toiIndexB) {
				mB = pc.invMassB
				iB = pc.invIB
			}

			var cA = positions[indexA].c
			var aA = positions[indexA].a

			var cB = positions[indexB].c
			var aB = positions[indexB].a

			// Solve normal constraints
			for (j in 0 until pc.localPoints.size) {
				val xfA = Transform()
				val xfB = Transform()

				xfA.q.set(aA)
				xfB.q.set(aB)

				xfA.p = cA - b2Mul(xfA.q, localCenterA)
				xfB.p = cB - b2Mul(xfB.q, localCenterB)

				val (normal, point, separation) = PositionSolverManifold(pc, xfA, xfB, j)

				val rA = point - cA
				val rB = point - cB

				// Track max constraint error.
				minSeparation = b2Min(minSeparation, separation)

				// Prevent large corrections and allow slop.
				val C = b2Clamp(b2_toiBaumgarte * (separation + b2_linearSlop), -b2_maxLinearCorrection, 0.0)

				// Compute the effective mass.
				val rnA = b2Cross(rA, normal)
				val rnB = b2Cross(rB, normal)
				val K = mA + mB + iA * rnA * rnA + iB * rnB * rnB

				// Compute normal impulse
				val impulse = if (K > 0.0) -C / K else 0.0

				val P = impulse * normal

				cA -= mA * P
				aA -= iA * b2Cross(rA, P)

				cB += mB * P
				aB += iB * b2Cross(rB, P)
			}

			positions[indexA].c = cA
			positions[indexA].a = aA

			positions[indexB].c = cB
			positions[indexB].a = aB
		}

		// We can't expect minSpeparation >= -b2_linearSlop because we don't
		// push the separation above -b2_linearSlop.
		return minSeparation >= -1.5f * b2_linearSlop
	}
}
