package ru.dbotthepony.kbox2d.dynamics.joint

import ru.dbotthepony.kbox2d.api.*
import ru.dbotthepony.kbox2d.api.B2SolverData
import ru.dbotthepony.kbox2d.api.b2Mul
import ru.dbotthepony.kvector.vector.ndouble.Vector2d
import ru.dbotthepony.kvector.vector.ndouble.times

/**
 * The pulley joint is connected to two bodies and two fixed ground points.
 * The pulley supports a ratio such that:
 * length1 + ratio * length2 <= constant
 * Yes, the force transmitted is scaled by the ratio.
 * Warning: the pulley joint can get a bit squirrelly by itself. They often
 * work better when combined with prismatic joints. You should also cover the
 * the anchor points with static shapes to prevent one side from going to
 * zero length.
 */
class PulleyJoint(def: PulleyJointDef) : AbstractJoint(def) {
	/**
	 * Get the first ground anchor.
	 */
	var groundAnchorA = def.groundAnchorA
		private set

	/**
	 * Get the second ground anchor.
	 */
	var groundAnchorB = def.groundAnchorB
		private set

	/**
	 * Get the current length of the segment attached to bodyA.
	 */
	val lengthA = def.lengthA

	/**
	 * Get the current length of the segment attached to bodyB.
	 */
	val lengthB = def.lengthB

	// Solver shared
	private val localAnchorA: Vector2d = def.localAnchorA
	private val localAnchorB: Vector2d = def.localAnchorB
	private val constant: Double = def.lengthA + def.ratio * def.lengthB

	/**
	 * Get the pulley ratio.
	 */
	val ratio: Double = def.ratio

	init {
		require(ratio != 0.0) { "Ratio is zero" }
	}

	private var impulse: Double = 0.0

	// Solver temp
	private var indexA: Int = 0
	private var indexB: Int = 0
	private var uA: Vector2d = Vector2d.ZERO
	private var uB: Vector2d = Vector2d.ZERO
	private var rA: Vector2d = Vector2d.ZERO
	private var rB: Vector2d = Vector2d.ZERO
	private var localCenterA: Vector2d = Vector2d.ZERO
	private var localCenterB: Vector2d = Vector2d.ZERO
	private var invMassA: Double = 0.0
	private var invMassB: Double = 0.0
	private var invIA: Double = 0.0
	private var invIB: Double = 0.0
	private var mass: Double = 0.0

	override fun initVelocityConstraints(data: B2SolverData) {
		indexA = bodyA.islandIndex
		indexB = bodyB.islandIndex
		localCenterA = bodyA.sweep.localCenter
		localCenterB = bodyB.sweep.localCenter
		invMassA = bodyA.invMass
		invMassB = bodyB.invMass
		invIA = bodyA.rotInertiaInv
		invIB = bodyB.rotInertiaInv

		val cA = data.positions[indexA].c
		val aA = data.positions[indexA].a
		var vA = data.velocities[indexA].v
		var wA = data.velocities[indexA].w

		val cB = data.positions[indexB].c
		val aB = data.positions[indexB].a
		var vB = data.velocities[indexB].v
		var wB = data.velocities[indexB].w

		val qA = Rotation(aA)
		val qB = Rotation(aB)

		rA = b2Mul(qA, localAnchorA - localCenterA)
		rB = b2Mul(qB, localAnchorB - localCenterB)

		// Get the pulley axes.
		uA = cA + rA - groundAnchorA
		uB = cB + rB - groundAnchorB

		val lengthA = uA.length
		val lengthB = uB.length

		if (lengthA > 10.0 * b2_linearSlop) {
			uA *= 1.0 / lengthA
		} else {
			uA = Vector2d.ZERO
		}

		if (lengthB > 10.0 * b2_linearSlop) {
			uB *= 1.0 / lengthB
		} else {
			uB = Vector2d.ZERO
		}

		// Compute effective mass.
		val ruA = b2Cross(rA, uA)
		val ruB = b2Cross(rB, uB)

		val mA = invMassA + invIA * ruA * ruA
		val mB = invMassB + invIB * ruB * ruB

		mass = mA + ratio * ratio * mB

		if (mass > 0.0) {
			mass = 1.0 / mass
		}

		if (data.step.warmStarting) {
			// Scale impulses to support variable time steps.
			impulse *= data.step.dtRatio

			// Warm starting.
			val PA = -(impulse) * uA
			val PB = (-ratio * impulse) * uB

			vA += invMassA * PA
			wA += invIA * b2Cross(rA, PA)
			vB += invMassB * PB
			wB += invIB * b2Cross(rB, PB)
		} else {
			impulse = 0.0
		}

		data.velocities[indexA].v = vA
		data.velocities[indexA].w = wA
		data.velocities[indexB].v = vB
		data.velocities[indexB].w = wB
	}

	override fun solveVelocityConstraints(data: B2SolverData) {
		var vA = data.velocities[indexA].v
		var wA = data.velocities[indexA].w
		var vB = data.velocities[indexB].v
		var wB = data.velocities[indexB].w

		val vpA = vA + b2Cross(wA, rA)
		val vpB = vB + b2Cross(wB, rB)

		val Cdot = -b2Dot(uA, vpA) - ratio * b2Dot(uB, vpB)
		val impulse = -mass * Cdot
		this.impulse += impulse

		val PA = -impulse * uA
		val PB = -ratio * impulse * uB
		vA += invMassA * PA
		wA += invIA * b2Cross(rA, PA)
		vB += invMassB * PB
		wB += invIB * b2Cross(rB, PB)

		data.velocities[indexA].v = vA
		data.velocities[indexA].w = wA
		data.velocities[indexB].v = vB
		data.velocities[indexB].w = wB
	}

	override fun solvePositionConstraints(data: B2SolverData): Boolean {
		var cA = data.positions[indexA].c
		var aA = data.positions[indexA].a
		var cB = data.positions[indexB].c
		var aB = data.positions[indexB].a

		val qA = Rotation(aA)
		val qB = Rotation(aB)

		val rA = b2Mul(qA, localAnchorA - localCenterA)
		val rB = b2Mul(qB, localAnchorB - localCenterB)

		// Get the pulley axes.
		var uA = cA + rA - groundAnchorA
		var uB = cB + rB - groundAnchorB

		val lengthA = uA.length
		val lengthB = uB.length

		if (lengthA > 10.0 * b2_linearSlop) {
			uA *= 1.0 / lengthA
		} else {
			uA = Vector2d.ZERO
		}

		if (lengthB > 10.0 * b2_linearSlop) {
			uB *= 1.0 / lengthB
		} else {
			uB = Vector2d.ZERO
		}

		// Compute effective mass.
		val ruA = b2Cross(rA, uA)
		val ruB = b2Cross(rB, uB)

		val mA = invMassA + invIA * ruA * ruA
		val mB = invMassB + invIB * ruB * ruB

		var mass = mA + ratio * ratio * mB

		if (mass > 0.0f) {
			mass = 1.0f / mass
		}

		val C = constant - lengthA - ratio * lengthB
		val linearError = b2Abs(C)

		val impulse = -mass * C

		val PA = -impulse * uA
		val PB = -ratio * impulse * uB

		cA += invMassA * PA
		aA += invIA * b2Cross(rA, PA)
		cB += invMassB * PB
		aB += invIB * b2Cross(rB, PB)

		data.positions[indexA].c = cA
		data.positions[indexA].a = aA
		data.positions[indexB].c = cB
		data.positions[indexB].a = aB

		return linearError < b2_linearSlop
	}

	override val anchorA: Vector2d
		get() = bodyA.getWorldPoint(localAnchorA)
	override val anchorB: Vector2d
		get() = bodyB.getWorldPoint(localAnchorB)

	override fun getReactionForce(inv_dt: Double): Vector2d {
		return inv_dt * (impulse * uB)
	}

	override fun getReactionTorque(inv_dt: Double): Double {
		return 0.0
	}

	/**
	 * Get the current length of the segment attached to bodyA.
	 */
	val currentLengthA: Double get() {
		val p = bodyA.getWorldPoint(localAnchorA)
		val s = groundAnchorA
		val d = p - s
		return d.length
	}

	/**
	 * Get the current length of the segment attached to bodyB.
	 */
	val currentLengthB: Double get() {
		val p = bodyB.getWorldPoint(localAnchorB)
		val s = groundAnchorB
		val d = p - s
		return d.length
	}

	override fun shiftOrigin(newOrigin: Vector2d) {
		groundAnchorA -= newOrigin
		groundAnchorB -= newOrigin
	}
}
