
@file:Suppress("unused")

package ru.dbotthepony.kvector.matrix.ndouble

import ru.dbotthepony.kvector.api.*
import ru.dbotthepony.kvector.api.concrete.IMatrix3d
import ru.dbotthepony.kvector.narray.Double2Dimensional
import ru.dbotthepony.kvector.vector.ndouble.*

/**
 * Represents immutable concrete matrix with 3x3 dimensions, storing values as [Double]s.
 *
 * See [Matrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [Vector3d], generic interface is represented by [IMatrix3d].
 */
class Matrix3d : AbstractMatrixVd<Matrix3d>, IMatrix3d<Matrix3d> {
	constructor(
		c00: Double = 1.0, c01: Double = 0.0, c02: Double = 0.0,
		c10: Double = 0.0, c11: Double = 1.0, c12: Double = 0.0,
		c20: Double = 0.0, c21: Double = 0.0, c22: Double = 1.0,
	) : super(3, 3, Double2Dimensional(3, 3).also {
		it[0, 0] = c00
		it[1, 0] = c10
		it[2, 0] = c20

		it[0, 1] = c01
		it[1, 1] = c11
		it[2, 1] = c21

		it[0, 2] = c02
		it[1, 2] = c12
		it[2, 2] = c22
	})

	private constructor(memory: Double2Dimensional) : super(3, 3, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Double2Dimensional): Matrix3d {
		requireSizeEquals(memory)
		return Matrix3d(memory)
	}

	override val c00: Double get() = memory[0, 0]
	override val c10: Double get() = memory[1, 0]
	override val c20: Double get() = memory[2, 0]
	override val c01: Double get() = memory[0, 1]
	override val c11: Double get() = memory[1, 1]
	override val c21: Double get() = memory[2, 1]
	override val c02: Double get() = memory[0, 2]
	override val c12: Double get() = memory[1, 2]
	override val c22: Double get() = memory[2, 2]

	override val r00 by this::c00
	override val r01 by this::c10
	override val r02 by this::c20
	override val r10 by this::c01
	override val r11 by this::c11
	override val r12 by this::c21
	override val r20 by this::c02
	override val r21 by this::c12
	override val r22 by this::c22

	override val a11 by this::c00
	override val a21 by this::c01
	override val a31 by this::c02
	override val a12 by this::c10
	override val a22 by this::c11
	override val a32 by this::c12
	override val a13 by this::c20
	override val a23 by this::c21
	override val a33 by this::c22

	override val b11 by this::c00
	override val b21 by this::c10
	override val b31 by this::c20
	override val b12 by this::c01
	override val b22 by this::c11
	override val b32 by this::c21
	override val b13 by this::c02
	override val b23 by this::c12
	override val b33 by this::c22

	override val c0: Vector3d get() = Vector3d(c00, c01, c02)
	override val c1: Vector3d get() = Vector3d(c10, c11, c12)
	override val c2: Vector3d get() = Vector3d(c20, c21, c22)

	override val r0: Vector3d get() = Vector3d(r00, r01, r02)
	override val r1: Vector3d get() = Vector3d(r10, r11, r12)
	override val r2: Vector3d get() = Vector3d(r20, r21, r22)

	override fun translate(vector: Vector2d): Matrix3d {
		return rm(
			r00, r01, r02 + vector.x,
			r10, r11, r12 + vector.y,
			r20, r21, r22,
		)
	}

	override fun translateWithMultiplication(vector: Vector2d): Matrix3d {
		return rm(
			r00, r01, r02 + vector.x * r00 + vector.y * r01,
			r10, r11, r12 + vector.x * r10 + vector.y * r11,
			r20, r21, r22,
		)
	}

	override fun scale(vector: Vector3d): Matrix3d {
		return rm(
			r00 * vector.x, r01, r02,
			r10, r11 * vector.y, r12,
			r20, r21, r22 * vector.z
		)
	}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Double = 1.0, r01: Double = 0.0, r02: Double = 0.0,
			r10: Double = 0.0, r11: Double = 1.0, r12: Double = 0.0,
			r20: Double = 0.0, r21: Double = 0.0, r22: Double = 1.0,
		): Matrix3d {
			return Matrix3d(
				c00 = r00, c10 = r01, c20 = r02,
				c01 = r10, c11 = r11, c21 = r12,
				c02 = r20, c12 = r21, c22 = r22,
			)
		}

		/**
		 * Initializes new zero matrix.
		 *
		 * Since [Matrix3d] is immutable, it is encouraged you use [ZERO] property instead.
		 */
		fun zero(): Matrix3d {
			return Matrix3d(
				c00 = 0.0, c10 = 0.0, c20 = 0.0,
				c01 = 0.0, c11 = 0.0, c21 = 0.0,
				c02 = 0.0, c12 = 0.0, c22 = 0.0,
			)
		}

		/**
		 * All values of this matrix are zero
		 */
		@JvmField val ZERO = zero()

		/**
		 * Main diagonal of this matrix is 1s
		 */
		@JvmField val IDENTITY = Matrix3d()
	}
}

/**
 * Represents mutable concrete matrix with 3x3 dimensions, storing values as [Double]s.
 *
 * See [MutableMatrix4d] for documentation, with reflection of this class.
 *
 * Vectorized access use [MutableVector3d], generic interface is represented by [IMatrix3d].
 */
class MutableMatrix3d : AbstractMutableMatrixVd<MutableMatrix3d>, IMatrix3d<MutableMatrix3d>, IMutableSquareMatrix<MutableMatrix3d, Vector2d, Vector3d> {
	constructor(
		c00: Double = 1.0, c01: Double = 0.0, c02: Double = 0.0,
		c10: Double = 0.0, c11: Double = 1.0, c12: Double = 0.0,
		c20: Double = 0.0, c21: Double = 0.0, c22: Double = 1.0,
	) : super(3, 3, Double2Dimensional(3, 3).also {
		it[0, 0] = c00
		it[1, 0] = c10
		it[2, 0] = c20

		it[0, 1] = c01
		it[1, 1] = c11
		it[2, 1] = c21

		it[0, 2] = c02
		it[1, 2] = c12
		it[2, 2] = c22
	})

	private constructor(memory: Double2Dimensional) : super(3, 3, memory)

	override val flexible: Boolean = false

	override fun factorize(memory: Double2Dimensional): MutableMatrix3d {
		requireSizeEquals(memory)
		return MutableMatrix3d(memory)
	}

	override fun translateWithMultiplicationMut(vector: Vector2d): MutableMatrix3d {
		r02 += vector.x * r00 + vector.y * r01
		r12 += vector.x * r10 + vector.y * r11
		return this
	}

	override fun scaleMut(vector: Vector3d): MutableMatrix3d {
		r00 *= vector.x
		r11 *= vector.y
		r22 *= vector.z
		return this
	}

	override var c00: Double
		get() = memory[0, 0]
		set(value) { memory[0, 0] = value }
	override var c10: Double
		get() = memory[1, 0]
		set(value) { memory[1, 0] = value }
	override var c20: Double
		get() = memory[2, 0]
		set(value) { memory[2, 0] = value }
	override var c01: Double
		get() = memory[0, 1]
		set(value) { memory[0, 1] = value }
	override var c11: Double
		get() = memory[1, 1]
		set(value) { memory[1, 1] = value }
	override var c21: Double
		get() = memory[2, 1]
		set(value) { memory[2, 1] = value }
	override var c02: Double
		get() = memory[0, 2]
		set(value) { memory[0, 2] = value }
	override var c12: Double
		get() = memory[1, 2]
		set(value) { memory[1, 2] = value }
	override var c22: Double
		get() = memory[2, 2]
		set(value) { memory[2, 2] = value }

	override var r00 by this::c00
	override var r01 by this::c10
	override var r02 by this::c20
	override var r10 by this::c01
	override var r11 by this::c11
	override var r12 by this::c21
	override var r20 by this::c02
	override var r21 by this::c12
	override var r22 by this::c22

	override var a11 by this::c00
	override var a21 by this::c01
	override var a31 by this::c02
	override var a12 by this::c10
	override var a22 by this::c11
	override var a32 by this::c12
	override var a13 by this::c20
	override var a23 by this::c21
	override var a33 by this::c22

	override var b11 by this::c00
	override var b21 by this::c10
	override var b31 by this::c20
	override var b12 by this::c01
	override var b22 by this::c11
	override var b32 by this::c21
	override var b13 by this::c02
	override var b23 by this::c12
	override var b33 by this::c22

	override var translation: Vector2d
		get() = super.translation
		set(value) {
			r02 = value.x
			r12 = value.y
		}

	override var c0: Vector3d
		get() = object : MutableVector3d(c00, c01, c02) {
			override var x by this@MutableMatrix3d::c00
			override var y by this@MutableMatrix3d::c01
			override var z by this@MutableMatrix3d::c02
		}
		set(value) {
			c00 = value.x
			c01 = value.y
			c02 = value.z
		}
	override var c1: Vector3d
		get() = object : MutableVector3d(c10, c11, c12) {
			override var x by this@MutableMatrix3d::c10
			override var y by this@MutableMatrix3d::c11
			override var z by this@MutableMatrix3d::c12
		}
		set(value) {
			c10 = value.x
			c11 = value.y
			c12 = value.z
		}
	override var c2: Vector3d
		get() = object : MutableVector3d(c20, c21, c22) {
			override var x by this@MutableMatrix3d::c20
			override var y by this@MutableMatrix3d::c21
			override var z by this@MutableMatrix3d::c22
		}
		set(value) {
			c20 = value.x
			c21 = value.y
			c22 = value.z
		}

	override var r0: Vector3d
		get() = object : MutableVector3d(r00, r01, r02) {
			override var x by this@MutableMatrix3d::r00
			override var y by this@MutableMatrix3d::r01
			override var z by this@MutableMatrix3d::r02
		}
		set(value) {
			r00 = value.x
			r01 = value.y
			r02 = value.z
		}
	override var r1: Vector3d
		get() = object : MutableVector3d(r10, r11, r12) {
			override var x by this@MutableMatrix3d::r10
			override var y by this@MutableMatrix3d::r11
			override var z by this@MutableMatrix3d::r12
		}
		set(value) {
			r10 = value.x
			r11 = value.y
			r12 = value.z
		}
	override var r2: Vector3d
		get() = object : MutableVector3d(r20, r21, r22) {
			override var x by this@MutableMatrix3d::r20
			override var y by this@MutableMatrix3d::r21
			override var z by this@MutableMatrix3d::r22
		}
		set(value) {
			r20 = value.x
			r21 = value.y
			r22 = value.z
		}

	override fun translate(vector: Vector2d): MutableMatrix3d {
		return rm(
			r00, r01, r02 + vector.x,
			r10, r11, r12 + vector.y,
			r20, r21, r22,
		)
	}

	override fun translateMut(vector: Vector2d): MutableMatrix3d {
		r02 += vector.x
		r12 += vector.y
		return this
	}

	override fun translateWithMultiplication(vector: Vector2d): MutableMatrix3d {
		return rm(
			r00, r01, r02 + vector.x * r00 + vector.y * r01,
			r10, r11, r12 + vector.x * r10 + vector.y * r11,
			r20, r21, r22,
		)
	}

	override fun scale(vector: Vector3d): MutableMatrix3d {
		return rm(
			r00 * vector.x, r01, r02,
			r10, r11 * vector.y, r12,
			r20, r21, r22 * vector.z
		)
	}

	companion object {
		/**
		 * Initializes new matrix, with arguments taken in *Row-Major* order, that is, fist row of values
		 * is fist row in created matrix.
		 */
		fun rm(
			r00: Double = 1.0, r01: Double = 0.0, r02: Double = 0.0,
			r10: Double = 0.0, r11: Double = 1.0, r12: Double = 0.0,
			r20: Double = 0.0, r21: Double = 0.0, r22: Double = 1.0,
		): MutableMatrix3d {
			return MutableMatrix3d(
				c00 = r00, c10 = r01, c20 = r02,
				c01 = r10, c11 = r11, c21 = r12,
				c02 = r20, c12 = r21, c22 = r22,
			)
		}

		/**
		 * Initializes new zero matrix.
		 */
		fun zero(): MutableMatrix3d {
			return MutableMatrix3d(
				c00 = 0.0, c10 = 0.0, c20 = 0.0,
				c01 = 0.0, c11 = 0.0, c21 = 0.0,
				c02 = 0.0, c12 = 0.0, c22 = 0.0,
			)
		}
	}
}

